# pylint: disable=simplifiable-if-expression
from abc import abstractmethod
from enum import Enum
from typing import List


class ExecutionState(str, Enum):
    disable = 'disable'
    waiting = 'waiting'
    running = 'running'
    done = 'done'


class Execution:

    def __init__(self, name: str, flow: 'Flow'):
        self._name = name
        flow.register_execution(self)
        self._depends = []  # type: List[Execution]
        self._status = ExecutionState.disable

    def depends(self, previous: 'Execution'):
        self._depends.append(previous)

    def name(self):
        return self._name

    def status(self):
        return self._status

    def initialize(self):
        self._status = ExecutionState.waiting

    def run_if_ready(self):
        for dependency in self._depends:
            dependency.check_if_done()

        if self._status == ExecutionState.waiting:
            ready_to_run = True
            for dependency in self._depends:
                ready_to_run &= False if dependency.status() != ExecutionState.done else True

            if ready_to_run:
                self.run()
                return True

        return False

    @abstractmethod
    def run(self):
        self._status = ExecutionState.running

    def check_if_done(self):
        if self.check():
            self._status = ExecutionState.done

    @abstractmethod
    def check(self) -> bool:
        raise NotImplementedError()



from subprocess import Popen

from oautom.execution.execution import Execution


class BashExecution(Execution):

    def run(self):
        super().run()
        self._proc = Popen(self._command, shell=True)

    def check(self) -> bool:
        check = False
        if self._proc:
            self._proc.poll()
            check = self._proc.returncode is not None

        return check

    def __init__(self, name, flow: 'Flow', command: str):
        super().__init__(name, flow)
        self._command = command
        self._proc = None  # type: Popen



# pylint: disable=useless-super-delegation
from oautom.execution.execution import Execution


class LambdaExecution(Execution):

    def __init__(self, name: str, flow: 'Flow'):
        super().__init__(name, flow)

    def run(self):
        raise NotImplementedError

    def check(self) -> bool:
        raise NotImplementedError



# pylint: disable=invalid-name

from flask import Flask, jsonify

from oautom.execution.bash_execution import BashExecution
from oautom import OAutom, Flow, OAutomMode

app = Flask(__name__)

oautom = OAutom(mode=OAutomMode.background)

flow1 = Flow('flow 1', app=oautom)
step1 = BashExecution('execution 1', flow=flow1, command='touch /tmp/file1')
step2 = BashExecution('sleep', flow=flow1, command='sleep 60')
step3 = BashExecution('execution 2', flow=flow1, command='touch /tmp/file2')
step2.depends(step1)
step3.depends(step2)
oautom.run()


@app.route("/flows")
def list_flow():
    return jsonify(oautom.flows())


@app.route("/logs")
def logs_flow():
    return jsonify(oautom.logs())


@app.route("/start/<flow>")
def start_flow(flow: str):
    result = ("OK", 200)
    if not oautom.start(flow):
        result = ("ALREADY RUNNING", 200)
    return result


@app.route("/status/<flow>")
def status_flow(flow: str):
    return jsonify(oautom.status(flow))


if __name__ == "__main__":
    app.run(debug=True)



from vect import Vect


class Flow:

    def __init__(self, name: str, app: 'OAutom'):
        self._name = name
        self._executions = []
        app.register_flow(self)

    def name(self):
        return self._name

    def register_execution(self, execution: 'Execution'):
        self._executions.append(execution)

    def vect(self):
        vect = Vect(self._name, self._executions)
        return vect



import json
from enum import Enum
from functools import partial
from typing import Dict

from apscheduler.schedulers.background import BackgroundScheduler
from apscheduler.schedulers.base import BaseScheduler
from apscheduler.schedulers.blocking import BlockingScheduler

from flow import Flow
from logger import get_logger
from vect import Vect


class OAutomMode(Enum):
    foreground = 'foreground'
    background = 'background'


class OAutom:

    def __init__(self, mode: OAutomMode = OAutomMode.foreground):
        self._logger = get_logger()
        self._flows = {}  # type: Dict[str, Flow]
        self._vects = {}  # type: Dict[str, Vect]
        self._logged_vects = []  # type: List[Vect]
        self._scheduler = self._event_loop(mode)

    def flows(self):
        return [flow.name() for flow in self._flows.values()]

    def logs(self):
        return [vect.status() for vect in self._logged_vects]

    def register_flow(self, flow: 'Flow'):
        self._flows[flow.name()] = flow

    def run(self):
        self._scheduler.start()

    def start(self, flow_name: str) -> bool:
        result = False
        vects = self._vects
        if flow_name not in vects or vects[flow_name].ended():
            vect = self._flows[flow_name].vect()
            vects[flow_name] = vect
            self._logged_vects.insert(0, vect)
            self._logger.info(f"run a new vect for {flow_name}")
            result = True

        return result

    def plan(self, flow_name: str, days: int = 0, hours: int = 0, minutes: int = 0):
        flow_vect = partial(self.start, flow_name)
        self._scheduler.add_job(flow_vect,
                                trigger="interval",
                                days=days,
                                hours=hours,
                                minutes=minutes)

    def status(self, flow_name: str):
        result = {}
        if flow_name in self._vects:
            result = self._vects[flow_name].status()

        return result

    def _event_loop(self, mode: OAutomMode) -> BaseScheduler:
        if mode == OAutomMode.background:
            scheduler = BackgroundScheduler()
        else:
            scheduler = BlockingScheduler()

        self._job = scheduler.add_job(func=self._move_vects_if_ready,
                                      trigger="interval",
                                      seconds=10)
        return scheduler

    def _move_vects_if_ready(self):
        for vect in self._vects.values():
            vect.run_if_ready()



import copy
from collections import OrderedDict
from datetime import datetime
from typing import List

from oautom.execution.execution import Execution, ExecutionState


class Vect:

    def __init__(self, name: str, executions: List[Execution]):
        executions = copy.deepcopy(executions)
        self._name = name
        self._executions = executions  # type: List[Execution]
        self._start_date = datetime.now()
        for execution in self._executions:
            execution.initialize()

    def run_if_ready(self):
        for execution in self._executions:
            execution.run_if_ready()
            execution.check_if_done()

    def ended(self) -> bool:
        ended = True
        for execution in self._executions:
            ended &= execution.status() == ExecutionState.done

        return ended

    def status(self) -> dict:
        status = OrderedDict()
        status['name'] = self._name
        status['ended'] = self.ended()
        status['start_date'] = self._start_date.isoformat()
        status['executions'] = []
        for execution in self._executions:
            status['executions'].append((execution.name(), execution.status()))

        return status



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from setuptools import setup
from setuptools import find_packages

setup(
    name='oautom',
    version='1.0.1',
    packages=find_packages(exclude=["*_tests"]),
    license='MIT',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url="https://github.com/FabienArcellier/oautom",
    install_requires = [
        'apscheduler'
    ],
    extras_require={
        'dev': [
            'pylint',
            'coverage',
            'flask',
            'twine'
        ]
    },
    classifier= [
        'Programming Language :: Python :: 3',
        'Framework :: Flask',
        'Operating System :: POSIX :: Linux'
    ],
)
